@@ -48,7 +48,7 @@ static const int error_connection = 1;
 
 static int set_fd_flags(const int fd)
 {
-    long flags = fcntl(fd, F_GETFL, 0);
+    int flags = fcntl(fd, F_GETFL, 0);
     if(flags == -1)
         return 0;
     flags |= O_NONBLOCK;
@@ -59,6 +59,21 @@ static int set_fd_flags(const int fd)
     return 1;
 }
 
+static int _xcb_xlib_init(_xcb_xlib *xlib)
+{
+    xlib->lock = 0;
+#ifndef NDEBUG
+    xlib->sloppy_lock = (getenv("LIBXCB_ALLOW_SLOPPY_LOCK") != 0);
+#endif
+    pthread_cond_init(&xlib->cond, 0);
+    return 1;
+}
+
+static void _xcb_xlib_destroy(_xcb_xlib *xlib)
+{
+    pthread_cond_destroy(&xlib->cond);
+}
+
 static int write_setup(xcb_connection_t *c, xcb_auth_info_t *auth_info)
 {
     static const char pad[3];
@@ -97,12 +112,12 @@ static int write_setup(xcb_connection_t *c, xcb_auth_info_t *auth_info)
     }
     assert(count <= sizeof(parts) / sizeof(*parts));
 
-    pthread_mutex_lock(&c->iolock);
+    _xcb_lock_io(c);
     {
         struct iovec *parts_ptr = parts;
         ret = _xcb_out_send(c, &parts_ptr, &count);
     }
-    pthread_mutex_unlock(&c->iolock);
+    _xcb_unlock_io(c);
     return ret;
 }
 
@@ -215,6 +230,7 @@ xcb_connection_t *xcb_connect_to_fd(int fd, xcb_auth_info_t *auth_info)
     if(!(
         set_fd_flags(fd) &&
         pthread_mutex_init(&c->iolock, 0) == 0 &&
+        _xcb_xlib_init(&c->xlib) &&
         _xcb_in_init(&c->in) &&
         _xcb_out_init(&c->out) &&
         write_setup(c, auth_info) &&
@@ -239,6 +255,7 @@ void xcb_disconnect(xcb_connection_t *c)
     close(c->fd);
 
     pthread_mutex_destroy(&c->iolock);
+    _xcb_xlib_destroy(&c->xlib);
     _xcb_in_destroy(&c->in);
     _xcb_out_destroy(&c->out);
 
@@ -255,15 +272,49 @@ void _xcb_conn_shutdown(xcb_connection_t *c)
     c->has_error = 1;
 }
 
+void _xcb_lock_io(xcb_connection_t *c)
+{
+    pthread_mutex_lock(&c->iolock);
+    while(c->xlib.lock)
+    {
+        if(pthread_equal(c->xlib.thread, pthread_self()))
+            break;
+        pthread_cond_wait(&c->xlib.cond, &c->iolock);
+    }
+}
+
+void _xcb_unlock_io(xcb_connection_t *c)
+{
+    pthread_mutex_unlock(&c->iolock);
+}
+
+void _xcb_wait_io(xcb_connection_t *c, pthread_cond_t *cond)
+{
+    int xlib_locked = c->xlib.lock;
+    if(xlib_locked)
+    {
+        c->xlib.lock = 0;
+        pthread_cond_broadcast(&c->xlib.cond);
+    }
+    pthread_cond_wait(cond, &c->iolock);
+    if(xlib_locked)
+    {
+        while(c->xlib.lock)
+            pthread_cond_wait(&c->xlib.cond, &c->iolock);
+        c->xlib.lock = 1;
+        c->xlib.thread = pthread_self();
+    }
+}
+
 int _xcb_conn_wait(xcb_connection_t *c, pthread_cond_t *cond, struct iovec **vector, int *count)
 {
-    int ret;
+    int ret, xlib_locked;
     fd_set rfds, wfds;
 
     /* If the thing I should be doing is already being done, wait for it. */
     if(count ? c->out.writing : c->in.reading)
     {
-        pthread_cond_wait(cond, &c->iolock);
+        _xcb_wait_io(c, cond);
         return 1;
     }
 
@@ -278,7 +329,13 @@ int _xcb_conn_wait(xcb_connection_t *c, pthread_cond_t *cond, struct iovec **vec
         ++c->out.writing;
     }
 
-    pthread_mutex_unlock(&c->iolock);
+    xlib_locked = c->xlib.lock;
+    if(xlib_locked)
+    {
+        c->xlib.lock = 0;
+        pthread_cond_broadcast(&c->xlib.cond);
+    }
+    _xcb_unlock_io(c);
     do {
 	ret = select(c->fd + 1, &rfds, &wfds, 0, 0);
     } while (ret == -1 && errno == EINTR);
@@ -287,7 +344,12 @@ int _xcb_conn_wait(xcb_connection_t *c, pthread_cond_t *cond, struct iovec **vec
         _xcb_conn_shutdown(c);
 	ret = 0;
     }
-    pthread_mutex_lock(&c->iolock);
+    _xcb_lock_io(c);
+    if(xlib_locked)
+    {
+        c->xlib.lock = 1;
+        c->xlib.thread = pthread_self();
+    }
 
     if(ret)
     {
